"""AMPQ support."""
import asyncio
import pickle
import uuid
from importlib import import_module

import aioamqp

from . import logger, AIOFALSE
from .utils import AsyncMixin, repr_func


class Queue(AsyncMixin):

    """Support message queue."""

    defaults = dict(
        host='localhost',
        port=None,
        login='guest',
        password='guest',
        virtualhost='/',
        queue='donald',
        exchange_name='',
    )

    def __init__(self, master, **params):
        """Initialize the queue."""
        self.params = self.defaults
        self.params.update(params)
        self.master = master
        self.transport = None
        self.protocol = None
        self.channel = None

        self._started = False
        self._connected = False
        self._loop = None

    def init_loop(self, loop):
        """Bind to given loop."""
        if not self._started:
            self._loop = loop or asyncio.get_event_loop()

    def is_connected(self):
        """Check that the queue is connected."""
        return self._connected

    async def start(self, listen=True, loop=None):
        """Connect and start listen the message queue."""
        if self.master.params.fake_mode:
            return True

        logger.warning('Start Donald Queue')

        self.init_loop(loop)
        self._started = True

        await self.connect()
        if listen:
            await self.channel.basic_consume(self.listen, queue_name=self.params['queue'])

    async def stop(self, *args, **kwargs):
        """Stop listeners."""
        if not self.is_connected() or self.is_closed() or not self._started:
            return False

        self._started = False

        await self.protocol.close(no_wait=True)
        self.transport.close()
        self._connected = False

        logger.warning('Donald Queue is stopped')

    async def __aenter__(self):
        """Support usage as a context manager."""
        await self.start()
        return self

    async def __aexit__(self, *args):
        """Support usage as a context manager."""
        await self.stop()

    async def connect(self):
        """Connect to queue."""
        if self._connected:
            return

        logger.warning('Connect to queue: %r', self.params)
        try:
            self.transport, self.protocol = await asyncio.wait_for(aioamqp.connect(
                loop=self._loop, on_error=self.on_error, **self.params), timeout=10
            )
            self.channel = await self.protocol.channel()
            await self.channel.queue_declare(queue_name=self.params['queue'], durable=True)
            await self.channel.basic_qos(
                prefetch_count=1, prefetch_size=0, connection_global=False)
            self._connected = True
        except Exception as exc:
            self.on_error(exc)
            raise

    def on_error(self, exc):
        """Error callback."""
        self._connected = False
        if not self._started or self.is_closed():
            return False

        logger.error(exc)
        self.loop.call_later(1, asyncio.create_task, self.connect())

    def submit(self, func, *args, **kwargs):
        """Submit to the queue."""
        if self.master.params.fake_mode:
            return self.master.submit(func, *args, **kwargs)

        logger.info('Submit to queue: %s', repr_func(func, args, kwargs))

        payload = pickle.dumps((func, args, kwargs))
        properties = dict(delivery_mode=2, message_id=str(uuid.uuid4()))

        if not self._connected:
            return AIOFALSE

        return asyncio.create_task(self.channel.basic_publish(
            payload=payload, exchange_name=self.params['exchange_name'],
            routing_key=self.params['queue'], properties=properties
        ))

    async def listen(self, channel, body, envelope, properties):
        """Run tasks from self.queue."""
        func, args, kwargs = pickle.loads(body)
        if isinstance(func, str):
            mod, _, func = func.rpartition('.')
            try:
                mod = import_module(mod)
                func = getattr(mod, func)
            except (ImportError, AttributeError) as exc:
                logger.error(exc)
                return False

        self.master.submit(func, *args, **kwargs)
        if channel:
            await channel.basic_client_ack(delivery_tag=envelope.delivery_tag)



"""Donald tasks' manager for Asyncio."""
import logging
import asyncio


AIOFALSE = asyncio.Future()
AIOFALSE.set_result(False)
AIOTRUE = asyncio.Future()
AIOTRUE.set_result(True)


logger = logging.getLogger(__name__)
handle = logging.StreamHandler()
handle.setFormatter(logging.Formatter(
    '[%(asctime)-15s] [%(process)d] [%(levelname)s] %(message)s', '%Y-%m-%d %H:%M:%S'))
logger.addHandler(handle)


from .core import Donald, CronTab as crontab, run_donald # noqa


__version__ = "0.9.1"
__project__ = "Donald"
__author__ = "Kirill Klenov <horneds@gmail.com>"
__license__ = "BSD"



"""CLI Support."""

from importlib import import_module

import click

from .core import Donald, run_donald


@click.command()
@click.option('--listen', help="Listen the given AMPQ URL.")
@click.option('--workers', default=Donald.defaults['num_workers'], help="Number of workers.")
@click.option('--loglevel', default=Donald.defaults['loglevel'], help="Logging level.")
@click.option('--module', help="Load module.")
def main(workers=4, loglevel='info', module=None, **kw):
    """Run Donald server."""
    donald = Donald(num_workers=workers, loglevel=loglevel)

    if module:
        try:
            click.echo('Import module: %s' % module)
            import_module(module)
        except ImportError:
            click.echo('Invalid module: %s' % module)

    run_donald(donald)


if __name__ == '__main__':
    main()



"""Donald Asyncio Tasks."""

import asyncio
import datetime
import multiprocessing as mp
import signal
from queue import Empty
from functools import wraps

from crontab import CronTab

from . import logger
from .queue import Queue
from .utils import AsyncMixin, AttrDict, FileLock, repr_func, create_task
from .worker import run_worker


class Donald(AsyncMixin):

    """I'am on Donald."""

    defaults = {

        # Run tasks imediatelly in the same process/thread
        'fake_mode': False,

        # Number of workers
        'num_workers': mp.cpu_count() - 1,

        # Maximum concurent tasks per worker
        'max_tasks_per_worker': 100,

        # Ensure that the Donald starts only once
        'filelock': None,

        # logging level
        'loglevel': 'INFO',

        # Start handlers
        'on_start': [],

        # Stop handlers
        'on_stop': [],

        # AMQP params
        'queue_name': 'donald',
        'queue_params': {},

        # Sentry options ({'dsn': '...'})
        'sentry': None,
    }

    crontab = CronTab
    workers = None
    listener = None
    queue = None

    _loop = None
    _started = False

    def __init__(self, on_start=None, on_stop=None, **params):
        """Initialize donald parameters."""
        self.params = AttrDict(self.defaults)
        self.params.update(params)

        logger.setLevel(self.params['loglevel'].upper())
        logger.propagate = False

        self.rx = self.tx = None

        self.lock = FileLock(self.params.filelock)
        self.schedules = []
        self.waiting = {}

        self.queue = Queue(self, **dict(self.params.queue_params, queue=self.params.queue_name))

    def __str__(self):
        """Representate as a string."""
        return f"Donald [{self.params.num_workers}]"

    async def start(self, loop=None):
        """Start workers.

        :returns: A coroutine
        """
        logger.warning('Start Donald: loop %s', id(asyncio.get_event_loop()))
        self._loop = loop or asyncio.get_event_loop()
        self.queue.init_loop(loop)

        if self.params.fake_mode:
            return True

        if self.params.filelock:
            try:
                self.lock.acquire()
            except self.lock.Error:
                logger.warning('Donald is locked. Exit.')
                return

        ctx = mp.get_context('spawn')
        self.rx = ctx.Queue()
        self.tx = ctx.Queue()

        self.workers = tuple(
            ctx.Process(target=run_worker, args=(self.rx, self.tx, self.params))
            for _ in range(max(self.params.num_workers, 1)))

        # Start workers
        for wrk in self.workers:
            wrk.start()
            self.tx.get(block=True)

        # Start listener
        self.listener = asyncio.create_task(self.listen())

        # Start schedulers
        for idx, schedule in enumerate(self.schedules):
            logger.info('Schedule %s', repr_func(schedule))
            self.schedules[idx] = asyncio.create_task(schedule())

        # Mark self started
        self._started = True

        return True

    async def stop(self, *args, **kwargs):
        """Stop workers. Disconnect from queue. Cancel schedules.

        The method could be called syncroniously too.

        :returns: A future
        """
        if self.is_closed() or not self._started:
            return

        logger.warning('Stoping Donald')

        if self.params.filelock:
            self.lock.release()

        # Stop runner if exists
        if self.listener:
            self.listener.cancel()

        # Stop schedules
        for task in self.schedules:
            task.cancel()

        # Stop workers
        for wrk in self.workers:
            self.rx.put(None)

        for wrk in self.workers:
            wrk.join(1)
            wrk.terminate()

        self.rx.close()
        self.tx.close()

        self._started = False

        if self.listener and not self.listener.done():
            await asyncio.sleep(1e-2)

        logger.warning('Donald is stopped')
        return True

    async def __aenter__(self):
        """Support usage as a context manager."""
        await self.start()
        return self

    async def __aexit__(self, *args):
        """Support usage as a context manager."""
        await self.stop()

    async def listen(self):
        """Wait for a result and process."""
        tx = self.tx
        waiting = self.waiting
        while True:
            if not waiting:
                await asyncio.sleep(1e-2)
                continue

            try:
                ident, res, *args = tx.get(block=False)
                fut = waiting.pop(ident, None)
                if fut:
                    if isinstance(res, Exception):
                        fut.set_exception(res)
                    else:
                        fut.set_result(res)

            except Empty:
                pass

            await asyncio.sleep(0)

    def submit(self, func, *args, **kwargs):
        """Submit a task to workers.

        :returns: asyncio.Future
        """
        if not callable(func):
            raise ValueError('Invalid call: %r' % func)

        if self.params.fake_mode:
            return create_task(func, args, kwargs)

        if not self._started:
            raise RuntimeError('Donald is not started yet')

        logger.debug('Submit: %s', repr_func(func, args, kwargs))
        fut = self.loop.create_future()
        fut_id = id(fut)
        self.waiting[fut_id] = fut
        self.rx.put((fut_id, func, args, kwargs))
        return fut

    def schedule(self, interval, *args, **kwargs):
        """Add func to schedules. Use this as a decorator.

        Run given func/coro periodically.
        """
        if callable(interval):
            raise RuntimeError('@donald.schedule(interval) should be used.')

        timer = lambda: float(interval) # noqa

        if isinstance(interval, datetime.timedelta):
            timer = interval.total_seconds

        elif isinstance(interval, CronTab):
            timer = lambda: interval.next(default_utc=True)  # noqa

        def wrapper(func):

            async def catcher():
                try:
                    await self.submit(func, *args, **kwargs)
                except Exception:
                    pass

            @wraps(func)
            async def scheduler():
                while True:
                    sleep = max(timer(), 0.01)
                    logger.info('Next %s in %0.2f s', repr_func(func, args, kwargs), sleep)
                    await asyncio.sleep(sleep)
                    asyncio.create_task(catcher())

            self.schedules.append(scheduler)
            return func

        return wrapper

    async def run(self, timer=60):
        """Keep asyncio busy."""
        while self._started:
            logger.info('Donald is running')
            await asyncio.sleep(timer)

    def on_start(self, func, *args, **kwargs):
        """Register start handler."""
        self.params.on_start.append((func, args, kwargs))
        return func

    def on_stop(self, func, *args, **kwargs):
        """Register stop handler."""
        self.params.on_stop.append((func, args, kwargs))
        return func


def run_donald(donald, queue=True):
    """Help to run donald."""
    loop = asyncio.get_event_loop()

    async def stop_donald():
        if queue:
            await donald.queue.stop()
        await donald.stop()
        loop.stop()

    def handle_signal(loop):
        """Stop donald before exit."""
        loop.remove_signal_handler(signal.SIGTERM)
        loop.remove_signal_handler(signal.SIGINT)
        loop.create_task(stop_donald())

    loop.add_signal_handler(signal.SIGTERM, handle_signal, loop)
    loop.add_signal_handler(signal.SIGINT, handle_signal, loop)
    loop.run_until_complete(donald.start())
    if queue:
        loop.run_until_complete(donald.queue.start())

    try:
        loop.run_forever()
    finally:
        loop.close()
